@@ -1,4 +1,4 @@
-from py65.utils.conversions import convert_to_bin, convert_to_bcd
+from py65.utils.conversions import convert_to_bin, convert_to_bcd, itoa
 from py65.utils.devices import make_instruction_decorator
 
 class MPU:
@@ -11,7 +11,7 @@ class MPU:
   NEGATIVE  = 128
   OVERFLOW  = 64
   UNUSED    = 32
-  BREAK     = 16
+  BREAK     = 16 # there is no BREAK flag, but this position indicates BREAK
   DECIMAL   = 8
   INTERRUPT = 4
   ZERO      = 2
@@ -36,9 +36,14 @@ class MPU:
     self.reset()
 
   def __repr__(self):
-    out = '<%s: A=%02x, X=%02x, Y=%02x, Flags=%02x, SP=%02x, PC=%04x>'
-    return out % (self.name, self.a, self.x, self.y,
-                  self.flags, self.sp, self.pc)
+    flags  = itoa(self.p, 2).rjust(8, '0')
+    indent = ' ' * (len(self.name) + 2)
+
+    out = "%s PC  AC XR YR SP NV-BDIZC\n" + \
+          "%s: %04x %02x %02x %02x %02x %s" 
+    
+    return out % (indent, self.name, 
+                  self.pc, self.a, self.x, self.y, self.sp, flags)
 
   def step(self):
     instructCode = self.ImmediateByte()
@@ -57,7 +62,7 @@ class MPU:
     self.a = 0
     self.x = 0
     self.y = 0
-    self.flags = self.UNUSED
+    self.p = self.p = self.BREAK | self.UNUSED
     self.processorCycles = 0
 
   # Helpers for addressing modes
@@ -68,6 +73,10 @@ class MPU:
   def WordAt(self, addr):
     return self.ByteAt(addr) + (self.ByteAt(addr + 1) << 8)
 
+  def WrapAt(self, addr):
+    wrap = lambda x: (x & 0xff00) + ((x + 1) & 0xff)
+    return self.ByteAt(addr) + (self.ByteAt(wrap(addr)) << 8)
+
   def ProgramCounter(self):
     return self.pc
 
@@ -86,17 +95,17 @@ class MPU:
     return 255 & (self.y + self.ByteAt(self.pc))
 
   def IndirectXAddr(self):
-    return self.WordAt( 255 & (self.ByteAt(self.pc) + self.x))
+    return self.WrapAt( 255 & (self.ByteAt(self.pc) + self.x))
 
   def IndirectYAddr(self):
     if self.addcycles:
-      a1 = self.WordAt(self.ByteAt(self.pc))
+      a1 = self.WrapAt(self.ByteAt(self.pc))
       a2 = (a1+self.y) & 0xffff
       if (a1 & 0xff00) != (a2 & 0xff00):
         self.excycles += 1
       return a2
     else:
-      return (self.WordAt(self.ByteAt(self.pc))+self.y)&0xffff
+      return (self.WrapAt(self.ByteAt(self.pc))+self.y)&0xffff
 
   def AbsoluteAddr(self):
     return self.WordAt(self.pc)
@@ -158,11 +167,11 @@ class MPU:
     return z
 
   def FlagsNZ(self, value):
-    self.flags &= ~(self.ZERO + self.NEGATIVE)
+    self.p &= ~(self.ZERO + self.NEGATIVE)
     if value == 0:
-      self.flags |= self.ZERO
+      self.p |= self.ZERO
     else:
-      self.flags |= value & self.NEGATIVE
+      self.p |= value & self.NEGATIVE
 
   # operations
 
@@ -177,16 +186,16 @@ class MPU:
       addr = x()
       tbyte = self.ByteAt(addr)
 
-    self.flags &= ~(self.CARRY + self.NEGATIVE + self.ZERO)
+    self.p &= ~(self.CARRY + self.NEGATIVE + self.ZERO)
 
     if tbyte & 128:
-      self.flags |= self.CARRY
+      self.p |= self.CARRY
     tbyte = (tbyte << 1) & 0xFF
 
     if tbyte:
-      self.flags |= tbyte & 128
+      self.p |= tbyte & 128
     else:
-      self.flags |= self.ZERO
+      self.p |= self.ZERO
     
     if x is None:
       self.a = tbyte
@@ -200,14 +209,14 @@ class MPU:
       addr = x()
       tbyte = self.ByteAt(addr)
 
-    self.flags &=~(self.CARRY+self.NEGATIVE+self.ZERO)
-    self.flags |=tbyte&1
+    self.p &=~(self.CARRY+self.NEGATIVE+self.ZERO)
+    self.p |=tbyte&1
 
     tbyte = tbyte >> 1
     if tbyte:
       pass # {}
     else:
-      self.flags |= self.ZERO
+      self.p |= self.ZERO
 
     if x is None:
       self.a = tbyte
@@ -215,22 +224,22 @@ class MPU:
       self.memory[addr]=tbyte
 
   def opBCL(self, x):
-    if self.flags & x: 
+    if self.p & x: 
       self.pc += 1
     else:
       self.BranchRelAddr()
 
   def opBST(self, x):
-    if self.flags & x:
+    if self.p & x:
       self.BranchRelAddr()
     else:
       self.pc += 1
 
   def opCLR(self, x):
-    self.flags &=~x
+    self.p &=~x
 
   def opSET(self, x):
-    self.flags |= x
+    self.p |= x
 
   def opAND(self, x):
     self.a &= self.ByteAt(x())
@@ -238,10 +247,10 @@ class MPU:
 
   def opBIT(self, x):
     tbyte = self.ByteAt(x())
-    self.flags &=~(self.ZERO+self.NEGATIVE+self.OVERFLOW)
+    self.p &=~(self.ZERO+self.NEGATIVE+self.OVERFLOW)
     if (self.a & tbyte) == 0:
-      self.flags |= self.ZERO
-    self.flags |= tbyte&(128+64)
+      self.p |= self.ZERO
+    self.p |= tbyte&(128+64)
 
   def opROL(self, x):
     if x is None:
@@ -250,15 +259,15 @@ class MPU:
       addr = x()
       tbyte = self.ByteAt(addr)
 
-    if self.flags & self.CARRY:
+    if self.p & self.CARRY:
       if tbyte & 128:
         pass
       else:
-        self.flags &= ~self.CARRY
+        self.p &= ~self.CARRY
       tbyte = (tbyte << 1) | 1
     else:
       if tbyte & 128:
-        self.flags |= self.CARRY
+        self.p |= self.CARRY
       tbyte = tbyte << 1     
     tbyte &= 0xFF
     self.FlagsNZ(tbyte)
@@ -275,39 +284,55 @@ class MPU:
   def opADC(self, x):
     data = self.ByteAt(x())
       
-    if self.flags & self.DECIMAL:
-      if self.flags & self.CARRY:
-        tmp = 1
-      else:
-        tmp = 0
-      data = convert_to_bin(data) + convert_to_bin(self.a) + tmp
-      self.flags &= ~(self.CARRY+self.OVERFLOW+self.NEGATIVE+self.ZERO)
-      if data > 99:
-        self.flags |= self.CARRY + self.OVERFLOW
-        data -= 100
+    if self.p & self.DECIMAL:
+      halfcarry = 0
+      decimalcarry = 0
+      adjust0 = 0
+      adjust1 = 0
+      nibble0 = (data & 0xf) + (self.a & 0xf) + (self.p & self.CARRY)
+      if nibble0 > 9:
+        adjust0 = 6
+        halfcarry = 1
+      nibble1 = ((data >> 4) & 0xf) + ((self.a >> 4) & 0xf) + halfcarry
+      if nibble1 > 9:
+        adjust1 = 6
+        decimalcarry = 1
 
-      if data == 0:
-        self.flags |= self.ZERO
+      # the ALU outputs are not decimally adjusted
+      nibble0 = nibble0 & 0xf
+      nibble1 = nibble1 & 0xf
+      aluresult = (nibble1 << 4) + nibble0
+
+      # the final A contents will be decimally adjusted
+      nibble0 = (nibble0 + adjust0) & 0xf
+      nibble1 = (nibble1 + adjust1) & 0xf
+      self.p &= ~(self.CARRY+self.OVERFLOW+self.NEGATIVE+self.ZERO)
+      if aluresult == 0:
+        self.p |= self.ZERO
       else:
-        self.flags |= data & 128
-      self.a = convert_to_bcd(data)
+        self.p |= aluresult & self.NEGATIVE
+      if decimalcarry == 1:
+        self.p |= self.CARRY
+      if ( ~(self.a ^ data) & (self.a ^ aluresult) ) & 0x80:
+        self.p |= self.OVERFLOW
+      self.a = (nibble1 << 4) + nibble0
     else:
-      if self.flags & self.CARRY:
+      if self.p & self.CARRY:
         tmp = 1
       else:
         tmp = 0
       result = data + self.a + tmp
-      self.flags &= ~(self.CARRY+self.OVERFLOW+self.NEGATIVE+self.ZERO)
+      self.p &= ~(self.CARRY+self.OVERFLOW+self.NEGATIVE+self.ZERO)
       if ( ~(self.a ^ data) & (self.a ^ result) ) & 0x80:
-        self.flags |= self.OVERFLOW
+        self.p |= self.OVERFLOW
       data = result
       if data > 255:
-        self.flags|=self.OVERFLOW+self.CARRY
+        self.p |= self.CARRY
         data &=255
       if data == 0:
-        self.flags |= self.ZERO
+        self.p |= self.ZERO
       else:
-        self.flags |= data & 128
+        self.p |= data & self.NEGATIVE
       self.a = data
 
   def opROR(self, x):
@@ -317,15 +342,15 @@ class MPU:
       addr=x()
       tbyte = self.ByteAt(addr)
 
-    if self.flags & self.CARRY:
+    if self.p & self.CARRY:
       if tbyte & 1:
         pass # {}
       else:
-        self.flags &=~ self.CARRY
+        self.p &=~ self.CARRY
       tbyte=(tbyte>>1)|128
     else:
       if tbyte & 1:
-        self.flags |= self.CARRY
+        self.p |= self.CARRY
       tbyte=tbyte>>1
     self.FlagsNZ(tbyte)
 
@@ -343,71 +368,69 @@ class MPU:
   def opSTX(self, y):
     self.memory[y()] = self.x
 
-  def opCPY(self, x):
-    tbyte=self.ByteAt(x())
-    self.flags &=~(self.CARRY+self.ZERO+self.NEGATIVE)
-    if self.y == tbyte:
-      self.flags |= self.CARRY + self.ZERO
-    elif self.y > tbyte:
-       self.flags |= self.CARRY
-    else:
-      self.flags |= self.NEGATIVE
-
-  def opCPX(self, y):
-    tbyte = self.ByteAt(y())
-    self.flags &=~(self.CARRY+self.ZERO+self.NEGATIVE)
-    if self.x == tbyte:
-      self.flags |= self.CARRY + self.ZERO
-    elif self.x > tbyte:
-      self.flags |= self.CARRY
-    else:
-      self.flags |= self.NEGATIVE
-
-  def opCMP(self, x):
-    tbyte = self.ByteAt(x())
-    self.flags &= ~(self.CARRY+self.ZERO+self.NEGATIVE)
-    if self.a == tbyte:
-      self.flags |= self.CARRY + self.ZERO
-    elif self.a > tbyte:
-      self.flags |= self.CARRY
-    else:
-      self.flags |= self.NEGATIVE
+  def opCMPR(self, get_address, register_value):
+    tbyte = self.ByteAt(get_address())
+    self.p &= ~(self.CARRY+self.ZERO+self.NEGATIVE)
+    if register_value == tbyte:
+      self.p |= self.CARRY + self.ZERO
+    elif register_value > tbyte:
+      self.p |= self.CARRY
+    self.p |= (register_value - tbyte) & self.NEGATIVE
 
   def opSBC(self, x):
     data = self.ByteAt(x())
       
-    if self.flags & self.DECIMAL:
-      if self.flags & self.CARRY:
-        borrow = 0
+    if self.p & self.DECIMAL:
+      halfcarry = 1
+      decimalcarry = 0
+      adjust0 = 0
+      adjust1 = 0
+
+      nibble0 = (self.a & 0xf) + (~data & 0xf) + (self.p & self.CARRY)
+      if nibble0 <= 0xf:
+        halfcarry = 0
+        adjust0 = 10
+      nibble1 = ((self.a >> 4) & 0xf) + ((~data >> 4) & 0xf) + halfcarry
+      if nibble1 <= 0xf:
+        adjust1 = 10 << 4
+
+      # the ALU outputs are not decimally adjusted
+      aluresult = self.a + (~data & 0xFF) + (self.p & self.CARRY)
+      if aluresult > 0xff:
+        decimalcarry = 1
+      aluresult &= 0xff
+
+      # but the final result will be adjusted
+      nibble0 = (aluresult + adjust0) & 0xf
+      nibble1 = ((aluresult + adjust1) >> 4) & 0xf
+
+      self.p &= ~(self.CARRY + self.ZERO + self.NEGATIVE + self.OVERFLOW)
+      if aluresult == 0:
+        self.p |= self.ZERO
       else:
-        borrow = 1
-    
-      data = convert_to_bin(self.a) - convert_to_bin(data) - borrow
-      self.flags &= ~(self.CARRY + self.ZERO + self.NEGATIVE + self.OVERFLOW)
-      if data == 0:
-        self.flags |= self.ZERO + self.CARRY
-      elif data > 0:
-        self.flags |= self.CARRY
-      else:
-        self.flags |= self.NEGATIVE
-        data +=100
-      self.a = convert_to_bcd(data)
+        self.p |= aluresult & self.NEGATIVE
+      if decimalcarry == 1:
+        self.p |= self.CARRY
+      if ( (self.a ^ data) & (self.a ^ aluresult) ) & 0x80:
+        self.p |= self.OVERFLOW
+      self.a = (nibble1 << 4) + nibble0
     else:
-      if self.flags & self.CARRY:
+      if self.p & self.CARRY:
         borrow = 0
       else:
         borrow = 1
 
-      data = self.a - data - borrow
-      self.flags &= ~(self.CARRY + self.ZERO + self.OVERFLOW + self.NEGATIVE)
+      result = self.a + (~data & 0xFF) + (self.p & self.CARRY)
+      self.p &= ~(self.CARRY + self.ZERO + self.OVERFLOW + self.NEGATIVE)
+      if ( (self.a ^ data) & (self.a ^ result) ) & 0x80:
+        self.p |= self.OVERFLOW
+      data = result & 0xFF
       if data == 0:
-        self.flags |= self.ZERO + self.CARRY
-      elif data > 0:
-        self.flags |= self.CARRY
-      else:
-        self.flags |= self.OVERFLOW
-      self.flags |= data & self.NEGATIVE
-      self.a = data & 0xFF
+        self.p |= self.ZERO
+      if result & 0x100:
+        self.p |= self.CARRY
+      self.p |= data & self.NEGATIVE
+      self.a = data
 
   def opDECR(self, x):
     if x is None:
@@ -416,12 +439,12 @@ class MPU:
       addr = x()
       tbyte = self.ByteAt(addr)
 
-    self.flags &= ~(self.ZERO + self.NEGATIVE)
+    self.p &= ~(self.ZERO + self.NEGATIVE)
     tbyte = (tbyte - 1) & 0xFF
     if tbyte:
-      self.flags |= tbyte & self.NEGATIVE
+      self.p |= tbyte & self.NEGATIVE
     else:
-      self.flags |= self.ZERO
+      self.p |= self.ZERO
 
     if x is None:
       self.a = tbyte
@@ -435,12 +458,12 @@ class MPU:
       addr = x()
       tbyte = self.ByteAt(addr)
 
-    self.flags &= ~(self.ZERO + self.NEGATIVE)
+    self.p &= ~(self.ZERO + self.NEGATIVE)
     tbyte = (tbyte + 1) & 0xFF
     if tbyte:
-      self.flags |= tbyte & self.NEGATIVE
+      self.p |= tbyte & self.NEGATIVE
     else:
-      self.flags |= self.ZERO
+      self.p |= self.ZERO
 
     if x is None:
       self.a = tbyte
@@ -479,10 +502,10 @@ class MPU:
     pc = (self.pc + 1) & 0xFFFF # The pc has already been increased one
     self.stPushWord(pc)
 
-    self.flags |= self.BREAK
-    self.stPush(self.flags)
+    self.p |= self.BREAK
+    self.stPush(self.p | self.BREAK | self.UNUSED)
 
-    self.flags |= self.INTERRUPT
+    self.p |= self.INTERRUPT
     self.pc = self.WordAt(self.IrqTo)
 
   @instruction(name="ORA", mode="inx", cycles=6)
@@ -502,7 +525,7 @@ class MPU:
   
   @instruction(name="PHP", mode="imp", cycles=3)
   def inst_0x08(self):
-    self.stPush(self.flags)
+    self.stPush(self.p | self.BREAK | self.UNUSED)
   
   @instruction(name="ORA", mode="imm", cycles=2)
   def inst_0x09(self):
@@ -588,7 +611,7 @@ class MPU:
 
   @instruction(name="PLP", mode="imp", cycles=4)
   def inst_0x28(self):
-    self.flags = self.stPop()
+    self.p = (self.stPop() | self.BREAK | self.UNUSED)
   
   @instruction(name="AND", mode="imm", cycles=2)
   def inst_0x29(self):
@@ -654,7 +677,7 @@ class MPU:
 
   @instruction(name="RTI", mode="imp", cycles=6)
   def inst_0x40(self):
-    self.flags = self.stPop()
+    self.p = (self.stPop() | self.BREAK | self.UNUSED)
     self.pc = self.stPopWord()
 
   @instruction(name="EOR", mode="inx", cycles=6)
@@ -774,7 +797,7 @@ class MPU:
   @instruction(name="JMP", mode="ind", cycles=5)
   def inst_0x6c(self):
     ta = self.WordAt(self.pc)
-    self.pc = self.WordAt(ta)
+    self.pc = self.WrapAt(ta)
 
   @instruction(name="ADC", mode="abs", cycles=4)
   def inst_0x6d(self):
@@ -1028,22 +1051,22 @@ class MPU:
 
   @instruction(name="CPY", mode="imm", cycles=2)
   def inst_0xc0(self):
-    self.opCPY(self.ProgramCounter)
+    self.opCMPR(self.ProgramCounter, self.y)
     self.pc += 1
 
   @instruction(name="CMP", mode="inx", cycles=6)
   def inst_0xc1(self):
-    self.opCMP(self.IndirectXAddr)
+    self.opCMPR(self.IndirectXAddr, self.a)
     self.pc += 1
 
   @instruction(name="CPY", mode="zpg", cycles=3)
   def inst_0xc4(self):
-    self.opCPY(self.ZeroPageAddr)
+    self.opCMPR(self.ZeroPageAddr, self.y)
     self.pc += 1
 
   @instruction(name="CMP", mode="zpg", cycles=3)
   def inst_0xc5(self):
-    self.opCMP(self.ZeroPageAddr)
+    self.opCMPR(self.ZeroPageAddr, self.a)
     self.pc += 1
 
   @instruction(name="DEC", mode="zpg", cycles=5)
@@ -1059,7 +1082,7 @@ class MPU:
 
   @instruction(name="CMP", mode="imm", cycles=2)
   def inst_0xc9(self):              
-    self.opCMP(self.ProgramCounter)
+    self.opCMPR(self.ProgramCounter, self.a)
     self.pc +=1
 
   @instruction(name="DEX", mode="imp", cycles=2)
@@ -1070,12 +1093,12 @@ class MPU:
 
   @instruction(name="CPY", mode="abs", cycles=4)
   def inst_0xcc(self):
-    self.opCPY(self.AbsoluteAddr)
+    self.opCMPR(self.AbsoluteAddr, self.y)
     self.pc += 2
 
   @instruction(name="CMP", mode="abs", cycles=4)
   def inst_0xcd(self):
-    self.opCMP(self.AbsoluteAddr)
+    self.opCMPR(self.AbsoluteAddr, self.a)
     self.pc += 2
 
   @instruction(name="DEC", mode="abs", cycles=3)
@@ -1089,12 +1112,12 @@ class MPU:
 
   @instruction(name="CMP", mode="iny", cycles=5, extracycles=1)
   def inst_0xd1(self):
-    self.opCMP(self.IndirectYAddr)
+    self.opCMPR(self.IndirectYAddr, self.a)
     self.pc += 1
   
   @instruction(name="CMP", mode="zpx", cycles=4)
   def inst_0xd5(self):
-    self.opCMP(self.ZeroPageXAddr)
+    self.opCMPR(self.ZeroPageXAddr, self.a)
     self.pc += 1
 
   @instruction(name="DEC", mode="zpx", cycles=6)
@@ -1108,12 +1131,12 @@ class MPU:
 
   @instruction(name="CMP", mode="aby", cycles=4, extracycles=1)
   def inst_0xd9(self):
-    self.opCMP(self.AbsoluteYAddr)
+    self.opCMPR(self.AbsoluteYAddr, self.a)
     self.pc += 2
 
   @instruction(name="CMP", mode="abx", cycles=4, extracycles=1)
   def inst_0xdd(self):
-    self.opCMP(self.AbsoluteXAddr)
+    self.opCMPR(self.AbsoluteXAddr, self.a)
     self.pc += 2
 
   @instruction(name="DEC", mode="abx", cycles=7)
@@ -1123,7 +1146,7 @@ class MPU:
 
   @instruction(name="CPX", mode="imm", cycles=2)
   def inst_0xe0(self):
-    self.opCPX(self.ProgramCounter)
+    self.opCMPR(self.ProgramCounter, self.x)
     self.pc += 1
 
   @instruction(name="SBC", mode="inx", cycles=6)
@@ -1133,7 +1156,7 @@ class MPU:
   
   @instruction(name="CPX", mode="zpg", cycles=3)
   def inst_0xe4(self):
-    self.opCPX(self.ZeroPageAddr)
+    self.opCMPR(self.ZeroPageAddr, self.x)
     self.pc += 1
   
   @instruction(name="SBC", mode="zpg", cycles=3)
@@ -1163,7 +1186,7 @@ class MPU:
 
   @instruction(name="CPX", mode="abs", cycles=4)
   def inst_0xec(self):
-    self.opCPX(self.AbsoluteAddr)
+    self.opCMPR(self.AbsoluteAddr, self.x)
     self.pc += 2
   
   @instruction(name="SBC", mode="abs", cycles=4)
