@@ -1,5 +1,10 @@
+from py65.utils.addressing import AddressParser
+
 class Disassembler:
-    def __init__(self, mpu, address_parser):
+    def __init__(self, mpu, address_parser=None):
+        if address_parser is None:
+            address_parser = AddressParser()
+
         self._mpu = mpu
         self._address_parser = address_parser
 
@@ -68,8 +73,8 @@ class Disassembler:
         elif addressing == 'rel':
             opv = self._mpu.ByteAt(pc + 1)
             targ = pc + 2
-            if opv & 128:
-                targ -= (opv ^ 255) + 1
+            if opv & (1<<(8-1)):
+                targ -= (opv ^ 0xFF) + 1
             else:
                 targ += opv
             targ &= 0xffff
